import numpy as np

class Configuration:
    """
    A :any:`Configuration` describes a specific arrangement of objects in the vector space of possible positions.
    Objects are represented by integers, with indistinguishable objects denoted by identical integers.
    This class subclasses `numpy.matrix <https://docs.scipy.org/doc/numpy/reference/generated/numpy.matrix.html>`_. 
    Each configuration in the vector space of positions is represented as a column vector.

    Attributes:
        count (int): If symmetry-inequivalent configurations have been generated for a `configuration space`,
                     this records the number of configurations equivalent to this one. 
                     Value at initialisation is  ``None``.
        lowest_numeric_representation (int): If the numeric representations for the set of equivalent 
                     configurations are calculated, this can be used to store the lowest valued numeric 
                     representation, for use as a simple hash.
    
    Example:

        >>> Configuration( [1, 1, 0] )
        Configuration([1, 1, 0])

    """ 

    def __init__( self, vector ):
        self.count = None
        self.lowest_numeric_representation = None
        self.vector = np.array( vector )

    def matches( self, test_configuration ):
        """
        Test whether this configuration is equal to another configuration.

        Args:
            test_configuration (:any:`Configuration`): The configuration to compare against.

        Returns:
            (bool): True | False.
        """
        if not isinstance( test_configuration, Configuration ):
            raise TypeError
        return ( self.vector == test_configuration.vector ).all()

    def is_equivalent_to( self, test_configuration, symmetry_operations ):
        """
        Test whether this configuration is equivalent to another configuration
        under one or more of a set of symmetry operations.

        Args:
            test_configuration (Configuration): The configuration to compare against.
            symmetry_operations (list(SymmetryOperation): A list of SymmetryOperation objects.

        Returns:
            (bool): True | False
        """
        for symmetry_operation in symmetry_operations:
            if ( symmetry_operation.operate_on( self ).matches( test_configuration ) ):
                return True 
        else:
            return False

    def is_in_list( self, the_list ):
        """
        Test whether this configuration is in a list of configurations.

        Args:
            list (list(bsym.Comfiguration)): A list of Configuration instances.

        Returns:
            (bool): True | False
        """
        return next( ( True for c in the_list if self.matches( c ) ), False )

    def has_equivalent_in_list( self, the_list, symmetry_operations ):
        """
        Test whether this configuration is equivalent by symmetry to one or more
        in a list of configurations.

        Args:
            list (list(bsym.Configuration)): A list of :any:`Configuration` instances.
            symmetry_operations (list(bsym.SymmetryOperation)): A list of :any:`SymmetryOperation` objects.

        Returns:
            (bool): True | False 
        """
        return next( ( True for c in the_list if self.is_equivalent_to( c, symmetry_operations ) ), False )

    def set_lowest_numeric_representation( self, symmetry_operations ):
       """
       Sets `self.lowest_numeric_representation` to the lowest value numeric representation of this configuration when permutated under a set of symmetry operations.

       Args:
           symmetry_operations (list): A list of :any:`SymmetryOperation` instances.

       Returns:
           None
       """
       self.lowest_numeric_representation = min( [ symmetry_operation.operate_on( self ).as_number for symmetry_operation in symmetry_operations ] )

    def numeric_equivalents( self, symmetry_operations ):
        """
        Returns a list of all symmetry equivalent configurations generated by a set of symmetry operations
        with each configuration given in numeric representation.

        Args:
            symmetry_operations (list): A list of :any:`SymmetryOperation` instances.

        Returns:
            (list(int)): A list of numbers representing each equivalent configuration.
        """
        return [ symmetry_operation.operate_on( self ).as_number for symmetry_operation in symmetry_operations ]
        #return [ as_number( symmetry_operation.operate_on( self, config=False ) ) for symmetry_operation in symmetry_operations ]

    @property
    def as_number( self ):
        """
        A numeric representation of this configuration.

        Examples:
            >>> c = Configuration( [ 1, 2, 0 ] )
            >>> c.as_number
            120

        """
        return as_number( self.vector )

    @classmethod
    def from_tuple( cls, this_tuple ):
        """
        Construct a :any:`Configuration` from a `tuple`,
        e.g.::
     
            Configuration.from_tuple( ( 1, 1, 0 ) )
 
        Args:
            this_tuple (tuple): The tuple used to construct this :any:`Configuration`.

        Returns:
            (:any:`Configuration`): The new :any:`Configuration`.
        """
        return( cls( this_tuple ) )

    def tolist( self ):
        """
        Returns the configuration data as a list.
        
        Args:
            None

        Returns:
            (List)
        """
        return list( self.vector )

    def pprint( self ):
        print( ' '.join( [ str(e) for e in self.tolist() ] ) )

    def position( self, label ):
        """
        Returns the vector indices where elements are equal to `label`.

        Args:
            label (int): The label used to select the vector positions.

        Returns:
            (list): A list of all positions that match `label`.
        """
        return [ i for i,x in enumerate( self.tolist() ) if x == label ]

    def __repr__( self ):
        to_return = "Configuration({})\n".format(self.vector)
        return to_return

    def map_objects( self, objects ):
        """
        Returns a dict of objects sorted according to this configuration.

        Args:
            objects [list]: A list of objects.

        Returns:
            sorted_objects [dict]: A dictionary of sorted objects.
        """
        if len( objects ) != len( self.vector ):
            raise ValueError
        sorted_objects = {}
        for key in set( self.vector ):
            sorted_objects[key] = [ o for k, o in zip( self.vector, objects ) if k == key ]
        return sorted_objects 

def as_number( a ):
    tot = 0
    for num in a:
        tot *= 10
        tot += int( num )
    return tot




import warnings
import sys

warnings.warn( "\nYou are trying to import bsym.bsym, which is not compatible with this version of bsym.\nPlease check the updated API information at https://github.com/bjmorgan/bsym,\n or download a compatible 0.1.* release at https://github.com/bjmorgan/bsym/releases.\n" )

sys.exit()



from bsym import ConfigurationSpace
import numpy as np

class CoordinateConfigSpace( ConfigurationSpace ):
    """
    A :any:`CoordinateConfigSpace` object is a :any`ConfigurationSpace` that has an associated
    set of coordinates. Each vector in the configuration vector space has a corresponding coordinate.
    """

    def __init__( self, coordinates, symmetry_group=None, objects=None ):
        """
        Create a :any:`CoordinateConfigSpace` object.

        Args:
            coordinates (np.array): The set of coordinates that describe the vector space of this configuration space.
            symmetry_group (:any:`SymmetryGroup`): The set of symmetry operations describing the symmetries of this configuration space.

        Returns:
            None
        """
        if objects is None:
            # Create a set of objects to represent the coordinates.
            objects = np.arange( len( coordinates ) ) + 1
        super().__init__( objects, symmetry_group )
        self.coordinates = coordinates

    def unique_coordinates( self, site_distribution, verbose=False ):
        """
        Find the symmetry inequivalent coordinates for a given site occupation.

        Args:
            site_distribution (dict): A dictionary that defines the number of each object 
                                      to be arranged in this system.

                                      e.g. for a structure with four sites, with two occupied (denoted `1`)
                                      and two unoccupied (denoted `0`)::

                                          { 1 : 2, 2 : 1 }
            verbose (opt:default=False): Print verbose output.

        Returns:
            unique_coordinates (list[dict]): A list of dicts. Each dict describes the set of coordinates for each site type.
        """
        unique_configs = self.unique_configurations( site_distribution, verbose=verbose )
        unique_coordinates = [ u.map_objects( self.coordinates ) for u in unique_configs ]
        return unique_coordinates



from .symmetry_operation import SymmetryOperation
from .symmetry_group import SymmetryGroup
from .space_group import SpaceGroup
from .point_group import PointGroup
from .configuration import Configuration
from .configuration_space import ConfigurationSpace
from .coordinate_config_space import CoordinateConfigSpace

__version__ = '1.0.b4'




from bsym.permutations import flatten_list, unique_permutations
from bsym import Configuration, SymmetryGroup, SymmetryOperation
import numpy as np

class ConfigurationSpace:

    def __init__( self, objects, symmetry_group=None ):
        """
        Create a :any:`ConfigurationSpace` object.
  
        Args:
            objects (list): The set of objects that define the vector space of this configuration space.
            symmetry_group (:any:`SymmetryGroup`): The set of symmetry operations describing the symmetries of this configuration space.

        Returns:
            None
        """
        # Check that all properties have compatible dimensions
        dim = len( objects )
        self.objects = objects
        if symmetry_group:
            for so in symmetry_group.symmetry_operations:
                if so.matrix.shape[0] != dim:
                    raise ValueError
            self.symmetry_group = symmetry_group
        else:
            self.symmetry_group = SymmetryGroup( symmetry_operations=[ SymmetryOperation( np.identity( dim, dtype=int ), label='E' ) ] )

    def __repr__( self ):
        to_return = "ConfigurationSpace\n"
        to_return += self.objects.__repr__() + "\n"
        to_return += "\n".join( self.symmetry_group.__repr__().split("\n")[1:] )
        return to_return

    def unique_configurations( self, site_distribution, verbose=False ):
        """
        Find the symmetry inequivalent configurations for a given population of objects.

        Args:
            site_distribution (dict): A dictionary that defines the number of each object 
                                      to be arranged in this system.

                                      e.g. for a system with four sites, with two occupied (denoted `1`)
                                      and two unoccupied (denoted `0`)::

                                          { 1 : 2, 2 : 1 }
            verbose (opt:default=False): Print verbose output.

        Returns:
            unique_configurations (list): A list of :any:`Configuration` objects, for each symmetry 
                                          inequivalent configuration. 
        """
        if verbose:
            print( 'total number of sites: ' + str( sum( site_distribution.values() ) ) )
            print( 'using {:d} symmetry operations: '.format( len( self.symmetry_group.symmetry_operations ) ) )
        permutations = []
        working = True
        seen = set()
        unique_configurations = []
        s = flatten_list( [ [ key ] * site_distribution[ key ] for key in site_distribution ] )
        for new_permutation in unique_permutations( s ):
            if permutation_as_config_number( new_permutation) not in seen:
                config = Configuration.from_tuple( new_permutation )
                numeric_equivalents = set( config.numeric_equivalents( self.symmetry_group.symmetry_operations ) )
                config.count = len( numeric_equivalents )
                [ seen.add( i ) for i in numeric_equivalents ]
                unique_configurations.append( config )
                if verbose:
                    print( "found {:d}, screened {:d}".format( len( unique_configurations ), len( seen ) ) )
        if verbose:
            print( 'unique configurations: ' + str( len( unique_configurations ) ) )
        return( unique_configurations )

def permutation_as_config_number( p ):
    """
    A numeric representation of a numeric list.

    Example:
        >>> permutation_as_config_number( [ 1, 1, 0, 0, 1 ] )
        11001
    """
    tot = 0
    for num in p:
        tot *= 10
        tot += num
    return tot



import numpy as np
from bsym.configuration import Configuration

def is_square( m ):
    """
    Test whether a numpy matrix is square.

    Args:
        m (np.matrix): The matrix.

    Returns:
        (bool): True | False.
    """
    return m.shape[0] == m.shape[1]

class SymmetryOperation:
    """
    `SymmetryOperation` class.
    """

    def __init__( self, matrix, label=None ):
        """
        Initialise a `SymmetryOperation` object

        Args:
            matrix (numpy.matrix|numpy.ndarray|list): 1D vector as either a
            `numpy.matrix`, `numpy.ndarray`, or `list` containing the site mappings
            for this symmetry operation.
            label (default=None) (str): optional string label for this `SymmetryOperation` object.
        Raises:
            TypeError: if matrix is not `numpy.matrix`, `numpy.ndarray`, or `list`.

        Returns:
            self
        """
        if isinstance( matrix, np.matrix ):
            self.matrix = np.array( matrix )
        elif isinstance( matrix, np.ndarray ):
            self.matrix = np.array( matrix )
        elif isinstance( matrix, list):
            self.matrix = np.array( matrix )
        else:
            raise TypeError
        if not is_square( self.matrix ):
            raise ValueError
        self.label = label

    def __mul__( self, other ):
        """
        Multiply this `SymmetryOperation` matrix with another `SymmetryOperation`.

        Args:
            other (SymmetryOperation, Configuration): the other symmetry operation or configuration or matrix
            for the matrix multiplication self * other.

        Returns:
            (SymmetryOperation): a new `SymmetryOperation` instance with the resultant matrix.
            (Configuration): if `other` is a `Configuration`.
        """
        if isinstance( other, SymmetryOperation ):
            return( SymmetryOperation( self.matrix.dot( other.matrix ) ) )
        elif isinstance( other, Configuration ):
            return( self.operate_on( other ) )
        else:
            raise TypeError

    def invert( self, label=None ):
        """
        Invert this `SymmetryOperation` object.

        Args:
            None
 
        Returns:
            A new `SymmetryOperation` object corresponding to the inverse matrix operation.
        """
        return SymmetryOperation( np.linalg.inv( self.matrix ).astype( int ), label=label )

    @classmethod
    def from_vector( cls, vector, count_from_zero=False, label=None ):
        """
        Initialise a SymmetryOperation object from a vector of site mappings.

        Args:
            vector (list): vector of integers defining a symmetry operation mapping.
            count_from_zero (default = False) (bool): set to True if the site index counts from zero.
            label (default=None) (str): optional string label for this `SymmetryOperation` object.
   
        Returns:
            a new SymmetryOperation object
        """
        if not count_from_zero:
            vector = [ x - 1 for x in vector ]
        dim = len( vector )
        new_symmetry_operation = cls( np.zeros( ( dim, dim ), dtype=int ), label=label )
        for index, element in enumerate( vector ):
            new_symmetry_operation.matrix[ element, index ] = 1
        return new_symmetry_operation

    def similarity_transform( self, s ):
        """
        Generate the SymmetryOperation produced by a similarity transform S^{-1}.M.S

        Args:
            s: the symmetry operation or matrix S.

        Returns:
            the SymmetryOperation produced by the similarity transform
        """
        return( s.invert() * self * s )

    def operate_on( self, configuration ):
        """
        Return the Configuration generated by appliying this symmetry operation

        Args:
            configuration (Configuration): the configuration / occupation vector to operate on

        Returns:
            (Configuration): the new configuration obtained by operating on configuration with this symmetry operation. 
        """
        if not isinstance( configuration, Configuration ):
            raise TypeError
        return Configuration( self.matrix.dot( configuration.vector ) )

    def character( self ):
        """
        Return the character of this symmetry operation (the trace of `self.matrix`).

        Args:
            none

        Returns:
            np.trace( self.matrix )
        """
        return np.trace( self.matrix )

    def as_vector( self, count_from_zero=False ):
        """
        Return a vector representation of this symmetry operation

        Args:
            count_from_zero (default = False) (bool): set to True if the vector representation counts from zero
      
        Returns:
            a vector representation of this symmetry operation (as a list)
        """
        offset = 0 if count_from_zero else 1
        return( [ row.index( 1 ) + offset for row in self.matrix.T.tolist() ] )

    def set_label( self, label ):
        """
        Set the label for this symmetry operation.
  
        Args:
            label: label to set for this symmetry operation
        Returns:
            self 
        """
        self.label = label
        return( self )

    def pprint( self ):
        """
        Pretty print for this symmetry operation

        Args:
            None
        Returns:
            None
        """
        label = self.label if self.label else '---'
        print( label + ' : ' + ' '.join( [ str(e) for e in self.as_vector() ] ) )
        
    def __repr__( self ):
        label = self.label if self.label else '---'
        return 'SymmetryOperation\nlabel(' + label + ")\n" + self.matrix.__repr__()



import numpy as np
from bsym import SymmetryOperation 

class SymmetryGroup:
    """
    :any:`SymmetryGroup` class.

    A :any:`SymmetryGroup` object contains a set of :any:`SymmetryOperation` objects.

    e.g.::

        SymmetryGroup( symmetry_operations=[ s1, s2, s3 ] )

    where `s1`, `s2`, and `s3` are :any:`SymmetryOperation` objects.

    :any:`SymmetryGroup` objects can also be created from files using the class methods::

        SymmetryGroup.read_from_file( filename )

    and::

        SymmetryGroup.read_from_file_with_labels( filename )
    """

    class_str = 'SymmetryGroup'

    def __init__( self, symmetry_operations=[] ):
        """
        Create a :any:`SymmetryGroup` object.

        Args:
            symmetry_operations (list): A list of :any:`SymmetryOperation` objects.

        Returns:
            None
        """
        self.symmetry_operations = symmetry_operations

    @classmethod
    def read_from_file( cls, filename ):
        """
        Create a :any:`SymmetryGroup` object from a file.
       
        The file format should be a series of numerical mappings representing each symmetry operation.

        e.g. for a pair of equivalent sites::

            # example input file to define the spacegroup for a pair of equivalent sites
            1 2
            2 1

        Args:
            filename (str): Name of the file to be read in.

        Returns:
            spacegroup (SymmetryGroup)	
        """
        data = np.loadtxt( filename, dtype=int )
        symmetry_operations = [ SymmetryOperation.from_vector( row.tolist() ) for row in data ]
        return( cls( symmetry_operations = symmetry_operations ) )

    @classmethod
    def read_from_file_with_labels( cls, filename ):
        """
        Create a :any:`SymmetryGroup` object from a file, with labelled symmetry operations.

        The file format should be a series of numerical mappings representing each symmetry operation, prepended with a string that will be used as a label.

        e.g. for a pair of equivalent sites::

            # example input file to define the spacegroup for a pair of equivalent sites
            E  1 2
            C2 2 1

        Args:
            filename (str): Name of the file to be read in.

        Returns:
            spacegroup (SymmetryGroup)
        """
        data = np.genfromtxt( filename, dtype=str )
        labels = [ row[0] for row in data ]
        vectors = [ [ float(s) for s in row[1:] ] for row in data ]
        symmetry_operations = [ SymmetryOperation.from_vector( v ) for v in vectors ]
        [ so.set_label( l ) for (l, so) in zip( labels, symmetry_operations ) ]
        return( cls( symmetry_operations=symmetry_operations ) )

    def save_symmetry_operation_vectors_to( self, filename ):
        """
        Save the set of vectors describing each symmetry operation in this :any:`SymmetryGroup` to a file.

        Args:
            filename (str): Name of the file to save to.

        Returns:
            None
        """
        operation_list = []
        for symmetry_operation in self.symmetry_operations:
            operation_list.append( symmetry_operation.as_vector() )
        np.savetxt( filename, np.array( operation_list ), fmt='%i' )

    def extend( self, symmetry_operations_list ):
        """
        Extend the list of symmetry operations in this :any:`SymmetryGroup`.

        Args:
            symmetry_operations_list (list): A list of :any:`SymmetryOperation` objects.

        Returns:
            self (:any:`SymmetryGroup`)
        """
        self.symmetry_operations.extend( symmetry_operations_list )
        return self

    def append( self, symmetry_operation ):
        """
        Append a :any:`SymmetryOperation` to this :any:`SymmetryGroup`.

        Args:
            symmetry_operation (:any:`SymmetryOperation`): The :any:`SymmetryOperation` to add.

        Returns:
            self (:any:`SymmetryGroup`)
        """
        self.symmetry_operations.append( symmetry_operation )
        return self

    def by_label( self, label ):
        """
        Returns the :any:`SymmetryOperation` with a matching label.

        Args:
            label (str): The label identifying the chosen symmetry operation.

        Returns:
            (:any:`SymmetryOperation`): The symmetry operation that matches this label.
        """ 
        return next((so for so in self.symmetry_operations if so.label == label), None)

    @property
    def labels( self ):
        """
        A list of labels for each :any:`SymmetryOperation` in this spacegroup.

        Args:
            None

        Returns:
            (list): A list of label strings.
        """
        return [ so.label for so in self.symmetry_operations ] 

    def __repr__( self ):
        to_return = '{}\n'.format( self.__class__.class_str )
        for so in self.symmetry_operations:
            to_return += "{}\t{}\n".format( so.label, so.as_vector() )
        return to_return

    @property
    def size( self ):
        return len( self.symmetry_operations )




import os
from bsym import __version__

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = open('README.md').read()

config = {
    'description': 'A Basic Symmetry Module',
    'long_description': long_description,
    'author': 'Benjamin J. Morgan',
    'author_email': 'b.j.morgan@bath.ac.uk',
    'url': 'https://github.com/bjmorgan/bsym',
    'download_url': "https://github.com/bjmorgan/bsym/archive/%s.tar.gz" % (__version__),
    'author_email': 'b.j.morgan@bath.ac.uk',
    'version': __version__,
    'install_requires': [ 'numpy', 
			  'pymatgen', 
                          'coverage',
                          'codeclimate-test-reporter' ],
    'license': 'MIT',
    'packages': [ 'bsym' ],
    'scripts': [],
    'name': 'bsym'
}

setup(**config)



def flatten_list( this_list ):
    return [ item for sublist in this_list for item in sublist ]

def unique_permutations( seq ):
    """
    Yield only unique permutations of seq in an efficient way.

    A python implementation of Knuth's "Algorithm L", also known from the 
    std::next_permutation function of C++, and as the permutation algorithm 
    of Narayana Pandita.
   
    see http://stackoverflow.com/questions/12836385/how-can-i-interleave-or-create-unique-permutations-of-two-stings-without-recurs/12837695
    """
    # Precalculate the indices we'll be iterating over for speed
    i_indices = range(len(seq) - 1, -1, -1)
    k_indices = i_indices[1:]
    # The algorithm specifies to start with a sorted version
    seq = sorted(seq)
    while True:
        #yield list( seq )
        yield list( seq )
        # Working backwards from the last-but-one index,           k
        # we find the index of the first decrease in value.  0 0 1 0 1 1 1 0
        for k in k_indices:
            if seq[k] < seq[k + 1]:
                break
        else:
            # Introducing the slightly unknown python for-else syntax:
            # else is executed only if the break statement was never reached.
            # If this is the case, seq is weakly decreasing, and we're done.
            return
        # Get item from sequence only once, for speed
        k_val = seq[k]
        # Working backwards starting with the last item,           k     i
        # find the first one greater than the one at k       0 0 1 0 1 1 1 0
        for i in i_indices:
            if k_val < seq[i]:
                break
        # Swap them in the most efficient way
        (seq[k], seq[i]) = (seq[i], seq[k])                #       k     i
                                                           # 0 0 1 1 1 1 0 0
        # Reverse the part after but not                           k
        # including k, also efficiently.                     0 0 1 1 0 0 1 1
        seq[k + 1:] = seq[-1:k:-1]
